from __future__ import annotations

import os
import itertools
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Type, Generator, get_args, Dict, ClassVar, Any, Tuple, get_type_hints, Union, Callable, List

from google.cloud import datastore

from gcp_pilot import exceptions

DEFAULT_NAMESPACE = os.environ.get("GCP_DATASTORE_NAMESPACE", default=None)
DEFAULT_PK_FIELD = "id"
MAX_ITEMS_PER_OPERATIONS = 500  # Datastore cannot write more than 500 items per call


def _chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


@dataclass
class DoesNotExist(Exception):
    cls: Type[EmbeddedDocument]
    filters: Dict


@dataclass
class MultipleObjectsFound(Exception):
    cls: Type[EmbeddedDocument]
    filters: Dict


def _starts_with_operator(lookup_fields, value) -> List[Tuple[str, str, Any]]:
    field_name = ".".join(lookup_fields)
    return [
        (field_name, ">=", value),
        (field_name, "<=", f"{value}\ufffd"),
    ]


@dataclass
class Manager:
    lookup_operators: ClassVar[Dict[str, Union[str, Callable]]] = {
        "eq": "=",
        "gt": ">",
        "gte": ">=",
        "lt": "<",
        "lte": "<=",
        "in": "in",
        "startswith": _starts_with_operator,
    }

    _client: ClassVar[datastore.Client] = None
    fields: Dict[str, type]
    pk_field: str
    doc_klass: Type[Document]
    kind: str

    def get_client(self) -> datastore.Client:
        if not self._client:
            self._client = datastore.Client(namespace=self.get_namespace())
        return self._client

    def get_namespace(self):
        return self.doc_klass.Meta.namespace

    def build_key(self, pk: Any = None) -> datastore.Key:
        if pk:
            typed_pk = self.doc_klass.Meta.fields[self.pk_field](pk)
            return self.get_client().key(self.kind, typed_pk)
        # If no primary key is provided, we let the server create a new ID
        return self.get_client().allocate_ids(self.get_client().key(self.kind), 1)[0]

    def _iterate(self, query, page_size):
        cursor = None
        empty = False

        while not empty:
            query_iter = query.fetch(start_cursor=cursor, limit=page_size)

            page = next(query_iter.pages, [])
            for item in page:
                yield item
            cursor = query_iter.next_page_token
            empty = not bool(cursor)

    def query(
        self,
        distinct_on: str = None,
        order_by: Union[str, List[str]] = None,
        page_size: int = None,
        **kwargs,
    ) -> datastore.query.Iterator:
        # base query
        query = self.get_client().query(kind=self.kind)
        if order_by:
            query.order = order_by
        if distinct_on:
            query.distinct_on = distinct_on

        # parse lookup args
        cross_filters = []
        for key, value in kwargs.items():
            all_filters = self._build_filter(key=key, value=value)
            for field_name, operator, field_value in all_filters:
                if operator == "in":
                    cross_filters.append((field_name, operator, field_value))
                else:
                    query.add_filter(field_name, operator, field_value)

        if not cross_filters:
            yield from self._iterate(query=query, page_size=page_size)
        else:
            # prepare combinations
            options = defaultdict(list)
            for name, operator, values in cross_filters:
                for value in values:
                    options[name].append((name, "=", value))

            combinations = itertools.product(*list(options.values()))

            found = set()
            for combination in combinations:
                current_query = query
                for field_name, operator, value in combination:
                    current_query = current_query.add_filter(field_name, operator, value)

                for item in self._iterate(query=current_query, page_size=page_size):
                    if item.id not in found:
                        yield item
                        found.add(item.id)

    def filter(self, **kwargs) -> Generator[Document, None, None]:
        for entity in self.query(**kwargs):
            yield self.from_entity(entity=entity)

    def get(self, **kwargs) -> Document:
        if self.pk_field in kwargs:
            pk = kwargs[self.pk_field]
            entity = self.get_client().get(key=self.build_key(pk=pk))
            if entity:
                return self.from_entity(entity=entity)
            raise DoesNotExist(self.doc_klass, pk)

        # Since we can't fetch directly from the key,
        # we filter and hope for just one object
        one_obj = None
        for obj in self.filter(**kwargs):
            if one_obj is not None:
                raise MultipleObjectsFound(self.doc_klass, filters=kwargs)
            one_obj = obj
        if not one_obj:
            raise DoesNotExist(self.doc_klass, filters=kwargs)
        return one_obj

    def create(self, obj: Document) -> Document:
        entity = self.to_entity(obj=obj)
        self.get_client().put(entity=entity)

        # if successfully saved, we assure the auto-generated ID is added to the final object
        if not obj.pk:
            setattr(obj, obj.Meta.pk_field, entity.id)
        return obj

    def update(self, pk: str, **kwargs) -> Document:
        if kwargs:
            entity = self.get_client().get(key=self.build_key(pk=pk))
            # TODO: enable partial nested updates
            as_data = {
                key: value.Meta.to_dict(obj=value) if isinstance(value, EmbeddedDocument) else value
                for key, value in kwargs.items()
            }
            entity.update(as_data)
            self.get_client().put(entity=entity)
        return self.get(id=pk)

    def delete(self, pk: str = None):
        if pk:
            self.get_client().delete(key=self.build_key(pk=pk))
        else:
            keys = [entity.key for entity in self.query()]
            for chunk in _chunks(keys, MAX_ITEMS_PER_OPERATIONS):
                self.get_client().delete_multi(keys=chunk)

    def _build_filter(self, key: str, value: Any) -> List[Tuple[str, str, Any]]:
        operator = None

        field_name, *extra = key.split("__")
        if extra:
            if len(extra) > 1:
                raise exceptions.UnsupportedFormatException(f"Unsupported lookup key format {extra}")
            operator = extra[0]
            if operator not in self.lookup_operators:
                raise exceptions.UnsupportedFormatException(f"Unsupported lookup {operator}")
            if callable(operator):
                return operator(field_name, value)

        parts = field_name.split(".")
        if len(parts) > 1 and parts[0] not in self.fields:
            raise exceptions.ValidationError(
                f"{parts[0]} is not a valid field. Excepted one of {' | '.join(self.fields)}"
            )

        return [(field_name, (operator or "="), value)]

    def to_entity(self, obj: Document) -> datastore.Entity:
        entity = datastore.Entity(key=self.build_key(pk=obj.pk))
        if not obj.pk:
            setattr(obj, obj.Meta.pk_field, entity.id)
        entity.update(obj.Meta.to_dict(obj=obj))
        return entity

    def from_entity(self, entity: datastore.Entity) -> Document:
        data = dict(entity.items())
        if self.pk_field not in data:
            data[self.pk_field] = entity.id
        return self.doc_klass.Meta.from_dict(data=data)


@dataclass
class Metadata:
    fields: Dict[str, type]
    doc_klass: Type[EmbeddedDocument]
    pk_field: str = None
    namespace: str = DEFAULT_NAMESPACE

    def from_dict(self, data: Dict) -> EmbeddedDocument:
        data = data.copy()

        def _build(klass: Union[EmbeddedDocument, Callable], value: Any):
            if value is None or klass == Any:  # pylint: disable=comparison-with-callable
                return value

            if issubclass(klass, EmbeddedDocument):
                return klass.Meta.from_dict(data=value)

            if klass == datetime:
                return klass.fromisoformat(str(value))

            return klass(value)

        parsed_data = {}
        for field_name, field_klass in self.fields.items():
            try:
                raw_value = data[field_name]
            except KeyError:
                continue

            if getattr(field_klass, "_name", "") == "List":
                inner_klass = get_args(field_klass)[0]  # TODO: test composite types
                item = [_build(klass=inner_klass, value=i) for i in raw_value]
            elif getattr(field_klass, "_name", "") == "Dict":
                inner_klass_key, inner_klass_value = get_args(field_klass)
                item = {
                    _build(klass=inner_klass_key, value=k): _build(klass=inner_klass_value, value=v)
                    for k, v in raw_value.items()
                }
            else:
                item = _build(klass=field_klass, value=raw_value)
            parsed_data[field_name] = item

        return self.doc_klass(**parsed_data)

    def to_dict(self, obj: EmbeddedDocument, select_fields: List[str] = None) -> dict:
        # TODO handle custom dynamic fields
        def _unbuild(value):
            if value is None:
                return value
            if isinstance(value, EmbeddedDocument):
                return value.Meta.to_dict(obj=value)
            if isinstance(value, Enum):
                return value.value
            return value

        data = {}
        for field_name, field_klass in self.fields.items():
            if select_fields and field not in select_fields:
                continue

            raw_value = getattr(obj, field_name)

            if getattr(field_klass, "_name", "") == "List":
                item = [_unbuild(value=i) for i in raw_value]
            elif getattr(field_klass, "_name", "") == "Dict":
                item = {_unbuild(value=k): _unbuild(value=v) for k, v in raw_value.items()}
            else:
                item = _unbuild(value=raw_value)

            data[field_name] = item

        return data


class ORM(type):
    def __new__(mcs, name, bases, attrs):
        is_abstract_model = mcs._is_abstract(name=name)
        is_concrete_model = mcs._is_concrete(bases=bases)

        if not is_abstract_model:
            # Since it was not explicitly provided, add id: str = None
            if not mcs._has_explicit_pk_field(attrs=attrs, bases=bases) and is_concrete_model:
                attrs["__annotations__"][DEFAULT_PK_FIELD] = int
                attrs[DEFAULT_PK_FIELD] = None

        new_cls = super().__new__(mcs, name, bases, attrs)

        if is_abstract_model:
            return new_cls

        # Metadata initialization
        typed_fields = mcs._extract_fields(klass=new_cls)
        new_cls.Meta = Metadata(
            fields=typed_fields,
            doc_klass=new_cls,
            namespace=getattr(new_cls, "__namespace__", None),
        )

        # Manager initialization
        if is_concrete_model:
            new_cls.documents = Manager(
                fields=typed_fields,
                pk_field=DEFAULT_PK_FIELD,
                doc_klass=new_cls,
                kind=name,
            )

            new_cls.Meta.pk_field = DEFAULT_PK_FIELD

        return new_cls

    @classmethod
    def _is_abstract(mcs, name: str) -> bool:
        return name in ["Document", "EmbeddedDocument"]

    @classmethod
    def _is_concrete(mcs, bases: Tuple[type]) -> bool:
        return "Document" in [base.__name__ for base in bases]

    @classmethod
    def _has_explicit_pk_field(mcs, attrs: Dict, bases: Tuple[type]) -> bool:
        if DEFAULT_PK_FIELD in attrs.get("__annotations__", []):
            return True

        for base in bases:
            if DEFAULT_PK_FIELD in get_type_hints(base):
                return True

        return False

    @classmethod
    def _extract_fields(mcs, klass: type) -> Dict[str, type]:
        def _ignore(var_type: str, k: type):
            is_private = var_type.startswith("_")
            is_class_var = getattr(k, "__origin__", None) == ClassVar  # pylint: disable=comparison-with-callable
            return is_private or is_class_var

        hints = get_type_hints(klass)
        typed_fields = {var_type: k for var_type, k in hints.items() if not _ignore(var_type, k)}
        return typed_fields


@dataclass
class EmbeddedDocument(metaclass=ORM):
    Meta: ClassVar[Metadata]

    @classmethod
    def deserialize(cls, **kwargs) -> EmbeddedDocument:
        return cls.Meta.from_dict(data=kwargs)

    def serialize(self) -> Dict:
        return self.Meta.to_dict(obj=self)


@dataclass
class Document(EmbeddedDocument):
    documents: ClassVar[Manager]

    @property
    def pk(self):
        return getattr(self, self.Meta.pk_field, None)

    def save(self) -> Document:
        return self.documents.create(obj=self)

    def delete(self) -> None:
        self.documents.delete(pk=self.id)


__all__ = (
    "DoesNotExist",
    "MultipleObjectsFound",
    "EmbeddedDocument",
    "Document",
)



import abc
import json
import logging
import os
from typing import Dict, Any, Callable, Tuple, List, Generator, Union

from google import auth
from google.auth import iam
from google.auth.credentials import Credentials
from google.auth.impersonated_credentials import Credentials as ImpersonatedCredentials
from google.auth.transport import requests
from google.oauth2.service_account import Credentials as ServiceAccountCredentials
from google.protobuf.duration_pb2 import Duration
from googleapiclient.discovery import build, Resource
from googleapiclient.errors import HttpError

from gcp_pilot import exceptions

DEFAULT_PROJECT = os.environ.get("GCP_PROJECT", None)
DEFAULT_LOCATION = os.environ.get("GCP_LOCATION", None)
DEFAULT_SERVICE_ACCOUNT = os.environ.get("GCP_SERVICE_ACCOUNT", None)

TOKEN_URI = "https://accounts.google.com/o/oauth2/token"

PolicyType = Dict[str, Any]
AuthType = Tuple[Credentials, str]
ImpersonatedAuthType = Tuple[ImpersonatedCredentials, str]
ResourceType = Dict[str, Any]

logger = logging.getLogger()

_CACHED_LOCATIONS = {}  # TODO: Implement a smarter solution for caching project's location


MINIMAL_SCOPES = [
    "https://www.googleapis.com/auth/cloud-platform",
]


class GoogleCloudPilotAPI(abc.ABC):
    _client_class = None
    _scopes: List[str] = []
    _iam_roles: List[str] = []
    _cached_credentials: AuthType = None
    _service_name = None
    _google_managed_service = False  # Service agent requires impersonation

    def __init__(
        self,
        subject: str = None,
        location: str = None,
        project_id: str = None,
        impersonate_account: str = None,
        credentials: Credentials = None,
        **kwargs,
    ):
        if credentials:
            self.credentials, credential_project_id = credentials, getattr(credentials, "project_id", None)
        else:
            self.credentials, credential_project_id = self._set_credentials(
                subject=subject,
                impersonate_account=impersonate_account or DEFAULT_SERVICE_ACCOUNT,
            )
        self.project_id = self._set_project_id(project_id=project_id, credential_project_id=credential_project_id)

        self.client = self._build_client(**kwargs)

        self.location = self._set_location(location=location)

    def _get_client_extra_kwargs(self):
        return {}

    def _build_client(self, **kwargs) -> Union[Resource, _client_class]:
        kwargs.update(self._get_client_extra_kwargs())

        return (self._client_class or build)(credentials=self.credentials, **kwargs)

    def _set_project_id(self, project_id: str, credential_project_id: str) -> str:
        return project_id or DEFAULT_PROJECT or credential_project_id

    def _set_location(self, location: str = None) -> str:
        return location or DEFAULT_LOCATION or self._get_project_default_location()

    @classmethod
    def _impersonate_account(
        cls,
        credentials: Credentials,
        service_account: str,
        scopes: List[str],
    ) -> ImpersonatedAuthType:
        credentials = ImpersonatedCredentials(
            source_credentials=credentials,
            target_principal=service_account,
            target_scopes=scopes,
            delegates=[],
        )

        # Fetch project from service account
        # Since we are impersonating this service account, use it's own project
        # TODO: regex this
        project_id = credentials.service_account_email.split("@")[-1].replace(".iam.gserviceaccount.com", "")

        return credentials, project_id

    @classmethod
    def _delegated_credential(
        cls,
        credentials: Credentials,
        subject: str,
        scopes: List[str],
    ) -> ServiceAccountCredentials:
        try:
            admin_credentials = credentials.with_subject(subject).with_scopes(scopes)
        except AttributeError:
            # When inside GCP, the credentials provided by the metadata service are immutable.
            # https://github.com/GoogleCloudPlatform/professional-services/tree/master/examples/gce-to-adminsdk

            request = requests.Request()
            credentials.refresh(request)

            # Create an IAM signer using the bootstrap credentials.
            signer = iam.Signer(
                request,
                credentials,
                credentials.service_account_email,
            )
            # Create OAuth 2.0 Service Account credentials using the IAM-based
            # signer and the bootstrap_credential's service account email.
            admin_credentials = ServiceAccountCredentials(
                signer,
                credentials.service_account_email,
                TOKEN_URI,
                scopes=scopes,
                subject=subject,
            )

        return admin_credentials

    @classmethod
    def _set_credentials(cls, subject: str = None, impersonate_account: str = None) -> AuthType:
        # Speed up consecutive authentications
        # TODO: check if this does not break multiple client usage
        all_scopes = MINIMAL_SCOPES + cls._scopes
        if not cls._cached_credentials:
            credentials, project_id = auth.default(scopes=all_scopes)
            cls._cached_credentials = credentials, project_id
        else:
            credentials, project_id = cls._cached_credentials

        if impersonate_account and getattr(credentials, "service_account_email") != impersonate_account:
            credentials, impersonated_project_id = cls._impersonate_account(
                credentials=credentials,
                service_account=impersonate_account,
                scopes=all_scopes,
            )
            project_id = impersonated_project_id or project_id

        if subject:
            credentials = cls._delegated_credential(credentials=credentials, subject=subject, scopes=all_scopes)

        return credentials, (project_id or getattr(credentials, "project_id", None))

    def get_oidc_token(self, audience: str = None) -> Dict[str, Dict[str, str]]:
        oidc_token = {"service_account_email": self.credentials.service_account_email}
        if audience:
            # TODO: make sure that, if URL, the query params are removed
            oidc_token["audience"] = audience
        return {"oidc_token": oidc_token}

    async def set_up_permissions(self, email: str, project_id: str = None) -> None:
        from gcp_pilot.resource import ResourceManager, ServiceAgent  # pylint: disable=import-outside-toplevel

        rm = ResourceManager()  # pylint: disable=invalid-name
        for role in self._iam_roles:
            await rm.add_member(
                email=email,
                role=role,
                project_id=project_id or self.project_id,
            )

        if self._google_managed_service:
            email = ServiceAgent.get_email(
                service_name=f"{self._service_name} Service Account",
                project_id=self.project_id,
            )

            await ResourceManager().allow_impersonation(
                email=email,
                project_id=project_id,
            )

    def _get_project_number(self, project_id: str) -> int:
        from gcp_pilot.resource import ResourceManager  # pylint: disable=import-outside-toplevel

        project = ResourceManager().get_project(project_id=project_id)
        return project["projectNumber"]

    def _as_duration(self, seconds) -> Duration:
        return Duration(seconds=seconds) if seconds else None

    @classmethod
    def build_from(cls, client: "GoogleCloudPilotAPI", project_id: str = None):
        return cls(
            credentials=client.credentials,
            project_id=project_id or client.project_id,
        )

    def _get_project_default_location(self, project_id: str = None) -> Union[str, None]:
        location = _CACHED_LOCATIONS.get(project_id or self.project_id, None)
        if location:
            return location

        from gcp_pilot.app_engine import AppEngine  # pylint: disable=import-outside-toplevel

        try:
            app_engine = AppEngine.build_from(client=self, project_id=project_id)
            return app_engine.location
        except exceptions.NotFound:
            return None

    def _project_path(self, project_id: str = None) -> str:
        return f"projects/{project_id or self.project_id}"

    def _location_path(self, project_id: str = None, location: str = None) -> str:
        project_path = self._project_path(project_id=project_id)
        return f"{project_path}/locations/{location or self.location}"


class AccountManagerMixin:
    def _as_member(self, email: str) -> str:
        if email == "allUsers":
            return email
        is_service_account = email.endswith(".gserviceaccount.com")
        prefix = "serviceAccount" if is_service_account else "member"
        return f"{prefix}:{email}"

    def _make_public(self, role: str, policy: Dict) -> Dict:
        return self._bind_email_to_policy(email="allUsers", role=role, policy=policy)

    def _make_private(self, role: str, policy: Dict) -> Dict:
        return self._unbind_email_from_policy(email="allUsers", role=role, policy=policy)

    def _bind_email_to_policy(self, email: str, role: str, policy: Dict) -> Dict:
        new_policy = policy.copy()

        role_id = role if (role.startswith("organizations/") or role.startswith("roles/")) else f"roles/{role}"
        member = self._as_member(email=email)

        try:
            binding = next(b for b in new_policy["bindings"] if b["role"] == role_id)
            if member in binding["members"]:
                return new_policy
            binding["members"].append(member)
        except (StopIteration, KeyError):
            binding = {"role": role_id, "members": [member]}

            new_bindings = new_policy.get("bindings", []).copy()
            new_bindings.append(binding)

            new_policy["bindings"] = new_bindings

        if "version" not in new_policy:
            new_policy["version"] = 1  # TODO: handle version 2 and 3 as its conditional roles
        return new_policy

    def _unbind_email_from_policy(self, email: str, role: str, policy: Dict):
        role_id = f"roles/{role}"
        member = self._as_member(email=email)

        try:
            binding = next(b for b in policy["bindings"] if b["role"] == role_id)
            if member not in binding["members"]:
                return policy
            binding["members"].remove(member)
        except (StopIteration, KeyError):
            pass
        return policy


class AppEngineBasedService:
    # This mixin must be used for all clients that uses GCP's resources that are based on App Engine
    # such as Cloud Scheduler.
    # They have a peculiar behaviour because the App Engine, when enabled the first time,
    # is created in a region and this region cannot be changed ever.
    # So these clients cannot just choose a desired region to work on, they must use the App Engine's
    # previously chosen region.
    def _set_location(self, location: str = None):
        project_location = self._get_project_default_location()

        explicit_location = location
        if explicit_location and explicit_location != project_location:
            logger.warning(
                f"Location {explicit_location} cannot be set in {self.__class__.__name__}. "
                f"It uses App Engine's default location for your project: {project_location}"
            )
        return project_location


def friendly_http_error(func):
    _reasons = {
        "notFound": exceptions.NotFound,
        "deleted": exceptions.AlreadyDeleted,
        "forbidden": exceptions.NotAllowed,
        "duplicate": exceptions.AlreadyExists,
        "push.webhookUrlUnauthorized": exceptions.PushWebhookInvalid,
    }
    _statuses = {
        "INVALID_ARGUMENT": exceptions.ValidationError,
        "PERMISSION_DENIED": exceptions.NotAllowed,
        "NOT_FOUND": exceptions.NotFound,
        "ALREADY_EXISTS": exceptions.AlreadyExists,
        "INVALID_PASSWORD": exceptions.InvalidPassword,
        "EMAIL_NOT_FOUND": exceptions.NotFound,
        "MISSING_ID_TOKEN": exceptions.MissingUserIdentification,
    }

    def inner_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HttpError as exc:
            errors = json.loads(exc.content)["error"]
            exception_klass = None
            details = ""

            if "errors" in errors:
                exception_klass = _reasons.get(errors["errors"][0]["reason"], None)

            if not exception_klass and "message" in errors:
                exception_klass = _statuses.get(errors["message"], None)

            if not exception_klass:
                exception_klass = _statuses.get(errors["status"], None)
                details = f"{errors['code']}: {errors['message']}"

            if exception_klass:
                raise exception_klass(details) from exc
            raise exc

    return inner_function


class DiscoveryMixin:
    @friendly_http_error
    def _execute(self, method: Callable, **kwargs) -> ResourceType:
        return method(**kwargs).execute()

    def _list(
        self,
        method: Callable,
        result_key: str = "items",
        params: Dict[str, Any] = None,
    ) -> Generator[ResourceType, None, None]:
        results = self._execute(
            method=method,
            **params,
        )
        for item in results.get(result_key, []):
            yield item

    def _paginate(
        self,
        method: Callable,
        result_key: str = "items",
        params: Dict[str, Any] = None,
        order_by: str = None,
        limit: int = None,
    ) -> Generator[ResourceType, None, None]:
        page_token = None
        params = params or {}

        if order_by:
            if order_by.startswith("-"):
                params["sortOrder"] = "DESCENDING"
                order_by = order_by[1:]
            params["orderBy"] = order_by

        if limit:
            params["maxResults"] = limit

        while True:
            results = self._execute(
                method=method,
                **params,
                pageToken=page_token,
            )
            for item in results.get(result_key, []):
                yield item

            page_token = results.get("nextPageToken")
            if not page_token:
                break



# More Information: <https://cloud.google.com/functions/docs/reference/rest>
from typing import Dict, Any, Generator

from gcp_pilot.base import GoogleCloudPilotAPI, DiscoveryMixin, PolicyType, AccountManagerMixin
from gcp_pilot import exceptions

FunctionType = Dict[str, Any]


class CloudFunctions(DiscoveryMixin, AccountManagerMixin, GoogleCloudPilotAPI):
    _scopes = [
        "https://www.googleapis.com/auth/cloudfunctions",
    ]

    def __init__(self, **kwargs):
        super().__init__(
            serviceName="cloudfunctions",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def _parent_path(self, project_id: str = None, location: str = None) -> str:
        return self._location_path(project_id=project_id, location=location)

    def _function_path(self, name: str, project_id: str = None, location: str = None) -> str:
        location_path = self._location_path(project_id=project_id, location=location)
        return f"{location_path}/functions/{name}"

    def get_functions(self, project_id: str = None, location: str = None) -> Generator[FunctionType, None, None]:
        params = dict(
            parent=self._parent_path(project_id=project_id, location=location),
        )
        yield from self._list(
            method=self.client.projects().locations().functions().list,
            result_key="functions",
            params=params,
        )

    def get_function(self, name: str, project_id: str = None, location: str = None) -> FunctionType:
        function_name = self._function_path(name=name, project_id=project_id, location=location)
        return self._execute(
            method=self.client.projects().locations().functions().get,
            name=function_name,
        )

    @classmethod
    def build_repo_source(
        cls,
        name,
        branch: str = None,
        commit: str = None,
        tag: str = None,
        directory: str = None,
        project_id: str = None,
    ) -> Dict[str, Any]:

        valid_references = [ref for ref in [branch, commit, tag] if ref]
        if len(valid_references) > 1:
            raise exceptions.ValidationError("Only one of branch|tag|commit must be provided")

        if branch:
            ref = f"moveable-aliases/{branch}"
        elif commit:
            ref = f"revisions{commit}"
        elif tag:
            ref = f"fixed-aliases/{tag}"
        else:
            ref = "moveable-aliases/master"

        path = f"/paths/{directory}" if directory else ""

        if "/" in name:
            # Assume it's a repository integrated through Github App
            org, repo = name.split("/")
            repo_name = f"github_{org}_{repo}"
        else:
            repo_name = name

        REPO_SOURCE_URL = "https://source.developers.google.com"  # pylint: disable=invalid-name
        url = f"{REPO_SOURCE_URL}/projects/{project_id}/repos/{repo_name}/{ref}{path}"
        repo = dict(url=url)

        return repo

    def create_or_update_function(
        self,
        name: str,
        description: str,
        entry_point: str,
        repo_name: str,
        runtime: str = "python39",
        timeout: int = 60,
        ram: int = 128,
        repo_branch: str = None,
        repo_tag: str = None,
        repo_commit: str = None,
        repo_directory: str = None,
        labels: Dict[str, str] = None,
        env_vars: Dict[str, str] = None,
        max_instances: int = None,
        project_id: str = None,
        location: str = None,
        service_account_email: str = None,
        is_public: bool = False,
    ) -> FunctionType:
        repo = self.build_repo_source(
            name=repo_name,
            commit=repo_commit,
            tag=repo_tag,
            branch=repo_branch,
            directory=repo_directory,
            project_id=project_id or self.project_id,
        )
        body = dict(
            name=self._function_path(name=name, project_id=project_id, location=location),
            description=description,
            entry_point=entry_point,
            runtime=runtime,
            labels=labels,
            environment_variables=env_vars,
            timeout=f"{timeout}s",
            available_memory_mb=ram,
            service_account_email=service_account_email,
            max_instances=max_instances,
            source_repository=repo,
            https_trigger=dict(security_level="SECURE_ALWAYS"),
        )

        try:
            fields_to_update = ",".join(body.keys())
            function_name = self._function_path(name=name, project_id=project_id, location=location)
            operation = self._execute(
                method=self.client.projects().locations().functions().patch,
                name=function_name,
                updateMask=fields_to_update,
                body=body,
            )
        except exceptions.NotFound:
            parent = self._parent_path(project_id=project_id, location=location)
            operation = self._execute(
                method=self.client.projects().locations().functions().create,
                location=parent,
                body=body,
            )

        self.set_permissions(name=name, project_id=project_id, location=location, is_public=is_public)

        return {"name": operation["metadata"]["target"]}

    def get_permissions(self, name: str, project_id: str = None, location: str = None) -> PolicyType:
        function_name = self._function_path(name=name, project_id=project_id, location=location)

        return self._execute(
            method=self.client.projects().locations().functions().getIamPolicy,
            resource=function_name,
        )

    def set_permissions(self, name: str, is_public: bool, project_id: str = None, location: str = None) -> PolicyType:
        policy = self.get_permissions(name=name, project_id=project_id, location=location)
        role = "roles/cloudfunctions.invoker"
        if is_public:
            new_policy = self._make_public(role=role, policy=policy)
        else:
            new_policy = self._make_private(role=role, policy=policy)

        function_name = self._function_path(name=name, project_id=project_id, location=location)
        body = dict(
            policy=new_policy,
            updateMask="bindings",
        )

        return self._execute(
            method=self.client.projects().locations().functions().setIamPolicy,
            resource=function_name,
            body=body,
        )

    def delete_function(self, name: str, project_id: str = None, location: str = None) -> FunctionType:
        function_name = self._function_path(name=name, project_id=project_id, location=location)
        operation = self._execute(
            method=self.client.projects().locations().functions().delete,
            name=function_name,
        )
        return {"name": operation["metadata"]["target"]}


__all__ = ("CloudFunctions",)



from datetime import datetime
from typing import Any, Dict, Union

from google.cloud import bigquery
from google.cloud.bigquery import Table, DatasetReference

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI
from gcp_pilot.storage import CloudStorage


class BigQuery(GoogleCloudPilotAPI):
    _client_class = bigquery.Client

    def _dataset_ref(self, dataset_name: str, project_id: str = None) -> DatasetReference:
        return self.client.dataset(
            dataset_id=dataset_name,
            project=project_id or self.project_id,
        )

    async def execute(
        self,
        sql: str,
        params: Dict[str, Any] = None,
        destination_table_name: str = None,
        destination_dataset_name: str = None,
        destination_project: str = None,
        truncate: bool = None,
    ):
        job_config = bigquery.QueryJobConfig()
        if destination_table_name or destination_dataset_name:
            if not destination_table_name and destination_dataset_name:
                raise exceptions.ValidationError(
                    "Both destination_dataset_name and destination_table_name must be provided."
                )
            destination_project = destination_project or self.project_id
            destination_dataset = destination_dataset_name
            destination = f"{destination_project}.{destination_dataset}.{destination_table_name}"
            job_config.destination = destination
            if truncate:
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
            elif truncate is False:
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND

        if params:
            query_params = [_BigQueryParam.parse(key, value) for key, value in params.items()]
            job_config.query_parameters = query_params

        query_job = self.client.query(sql, job_config=job_config)
        return self._wait_for_job(job=query_job)

    async def insert_rows(self, dataset_name: str, table_name: str, rows, project_id: str = None):
        table = await self.get_table(
            table_name=table_name,
            project_id=project_id,
            dataset_name=dataset_name,
        )
        errors = self.client.insert_rows(table=table, rows=rows)
        if errors:
            raise Exception(f"Bigquery insert error: {errors}")

    async def get_table(self, table_name: str, project_id: str = None, dataset_name: str = None) -> Table:
        dataset_ref = self._dataset_ref(project_id=project_id, dataset_name=dataset_name)
        table_ref = dataset_ref.table(table_id=table_name)
        return self.client.get_table(table_ref)

    async def load(
        self,
        table_name: str,
        filename: str,
        project_id: str = None,
        dataset_name: str = None,
        schema=None,
        wait: bool = False,
        truncate: bool = None,
        gcs_bucket: str = None,
    ) -> None:
        job_config = bigquery.LoadJobConfig()
        if schema:
            job_config.schema = schema
        else:
            job_config.autodetect = True
            job_config._properties["load"]["schemaUpdateOptions"] = ["ALLOW_FIELD_ADDITION"]

        extension = filename.split(".")[-1]
        if extension == "json":
            source_format = bigquery.job.SourceFormat.NEWLINE_DELIMITED_JSON
        elif extension == "csv":
            source_format = bigquery.job.SourceFormat.CSV
        else:
            message = "Unsupported BigQuery Source format {}".format(extension)
            raise exceptions.UnsupportedFormatException(message)

        job_config.source_format = source_format
        if truncate:
            job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
        elif truncate is False:
            job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND

        dataset_ref = self._dataset_ref(
            project_id=project_id or self.project_id,
            dataset_name=dataset_name,
        )
        target_ref = dataset_ref.table(table_name)

        is_gcs = filename.startswith("gs://")
        if not is_gcs:
            if gcs_bucket:
                gcs = CloudStorage(project_id=self.project_id)
                blob = await gcs.upload(
                    source_file=filename,
                    bucket_name=gcs_bucket,
                )
                file_url = gcs.get_uri(blob=blob)

                load_job = self.client.load_table_from_uri(
                    source_uris=file_url,
                    destination=target_ref,
                    job_config=job_config,
                )
            else:
                with open(filename, "rb") as file:
                    load_job = self.client.load_table_from_file(
                        file_obj=file,
                        destination=target_ref,
                        job_config=job_config,
                    )
        else:
            load_job = self.client.load_table_from_uri(
                source_uris=filename,
                destination=target_ref,
                job_config=job_config,
            )

        if wait:
            self._wait_for_job(job=load_job)

    async def copy(
        self,
        source_dataset_name: str,
        source_table_name: str,
        destination_table_name: str,
        destination_dataset_name: str,
        destination_project: str = None,
        wait: bool = False,
    ) -> None:
        source_ref = await self.get_table(dataset_name=source_dataset_name, table_name=source_table_name)
        dataset_ref = self._dataset_ref(
            project_id=destination_project or self.project_id,
            dataset_name=destination_dataset_name,
        )
        target_ref = dataset_ref.table(destination_table_name)

        job = self.client.copy_table(sources=source_ref, destination=target_ref)
        if wait:
            self._wait_for_job(job=job)

    @classmethod
    def date_to_str(cls, dt, table_suffix=False):
        if table_suffix:
            mask = "%Y%m%d"
        elif isinstance(dt, datetime):
            mask = "%Y-%m-%d %H:%M:%S"
        else:
            mask = "%Y-%m-%d"
        return dt.strftime(mask)

    def _wait_for_job(self, job):
        try:
            return job.result()
        except Exception as exc:
            raise exceptions.BigQueryJobError(job) from exc

    def add_external_gcs_source(
        self,
        gcs_url: str,
        dataset_name: str,
        table_name: str,
        skip_rows: int = 0,
        delimiter: str = ",",
        quote: str = '"',
        source_format: str = "CSV",
        project_id: str = None,
    ):
        dataset_ref = self._dataset_ref(dataset_name=dataset_name, project_id=project_id)
        table = dataset_ref.table(table_name)

        external_config = bigquery.ExternalConfig(source_format=source_format)
        external_config.autodetect = True
        external_config.source_uris = [gcs_url]
        external_config.options.skip_leading_rows = skip_rows
        external_config.options.field_delimiter = delimiter
        external_config.options.quote = quote

        table.external_data_configuration = external_config

        return self.client.create_table(table)


class _BigQueryParam:
    @classmethod
    def _get_type(cls, variable: Any) -> str:
        TYPES = {  # pylint: disable=invalid-name
            "int": "INT64",
            "str": "STRING",
            "datetime": "DATETIME",
            "date": "DATE",
            "bool": "BOOL",
            "float": "FLOAT64",
            "Decimal": "FLOAT64",
        }

        python_type = type(variable).__name__
        param_type = TYPES.get(python_type, None)
        if param_type is None:
            raise exceptions.ValidationError(f"Parameter with type {param_type} not supported")
        return param_type

    @classmethod
    def _get_value(cls, variable: Any) -> Any:
        TYPES = {  # pylint: disable=invalid-name
            "Decimal": float,
        }
        python_type = type(variable).__name__
        param_type = TYPES.get(python_type, None)
        if param_type is not None:
            return param_type(variable)
        return variable

    @classmethod
    def parse(cls, key: str, value: Any) -> Union[bigquery.ArrayQueryParameter, bigquery.ScalarQueryParameter]:
        if isinstance(value, list):
            param_type = cls._get_type(variable=value[0])
            param_class = bigquery.ArrayQueryParameter
        else:
            param_type = cls._get_type(variable=value)
            param_class = bigquery.ScalarQueryParameter
        return param_class(key, param_type, cls._get_value(value))


__all__ = ("BigQuery",)



# Reference <https://cloud.google.com/identity-platform/docs/apis>
import json
from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from typing import Dict, Optional, Union, Iterator
from urllib.parse import urlparse, parse_qs

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI, DiscoveryMixin, ResourceType


class OOBCodeType(Enum):
    RESET = "PASSWORD_RESET"
    VERIFY = "VERIFY_EMAIL"
    SIGNIN = "EMAIL_SIGNIN"


def parse_timestamp(timestamp: Union[str, int, float]) -> Optional[datetime]:
    if not timestamp:
        return None
    if len(str(timestamp)) >= 10:
        timestamp = float(timestamp) / 1000
    return datetime.fromtimestamp(float(timestamp), tz=timezone.utc)


@dataclass
class User:
    id: str
    email: str
    verified: bool
    disabled: bool
    created_at: Optional[datetime]
    last_login_at: Optional[datetime] = None
    password_changed_at: Optional[datetime] = None
    extra_attributes: Dict[str, str] = None

    @classmethod
    def create(cls, data: ResourceType) -> "User":
        return cls(
            id=data["localId"],
            email=data["email"],
            verified=data["emailVerified"],
            disabled=data["disabled"],
            created_at=parse_timestamp(timestamp=data["createdAt"]),
            last_login_at=parse_timestamp(timestamp=data.get("lastLoginAt")),
            password_changed_at=parse_timestamp(timestamp=data.get("passwordUpdatedAt")),
            extra_attributes=json.loads(data.get("customAttributes", "{}")),
        )


class IdentityPlatform(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = ["https://www.googleapis.com/auth/identitytoolkit"]

    def __init__(self, **kwargs):
        super().__init__(
            serviceName="identitytoolkit",
            version="v1",
            cache_discovery=False,
            static_discovery=False,
            **kwargs,
        )

    def find(self, email: str = None, phone_number: str = None, project_id: str = None) -> User:
        try:
            return next(self.lookup(email=email, phone_number=phone_number, project_id=project_id))
        except StopIteration as exc:
            raise exceptions.NotFound() from exc

    def lookup(self, email: str = None, phone_number: str = None, project_id: str = None) -> Iterator[User]:
        if not email and not phone_number:
            raise exceptions.ValidationError("Either `email` or `phone_number` must be provided")

        data = {
            "target_project_id": project_id or self.project_id,
        }
        if email:
            data["email"] = email
        if phone_number:
            data["phone_number"] = phone_number

        response = self._execute(method=self.client.accounts().lookup, body=data)
        for item in response.get("users", []):
            yield User.create(data=item)

    def sign_in_with_password(self, email: str, password: str):
        data = {
            "email": email,
            "password": password,
        }
        response = self._execute(method=self.client.accounts().signInWithPassword, body=data)
        return response

    def sign_in_with_phone_number(self, phone_number: str, code: str):
        data = {
            "phone_number": phone_number,
            "code": code,
        }
        response = self._execute(method=self.client.accounts().signInWithPhoneNumber, body=data)
        return response

    def sign_in_with_email_link(self, email: str, code: str):
        data = {
            "email": email,
            "oobCode": code,
        }
        response = self._execute(method=self.client.accounts().signInWithEmailLink, body=data)
        return response

    def generate_email_code(
        self,
        type: OOBCodeType,  # pylint: disable=redefined-builtin
        email: str,
        ip_address: str = None,
        project_id: str = None,
        send_email: bool = False,
        redirect_url: str = None,
    ) -> ResourceType:
        data = {
            "requestType": type.value,
            "email": email,
            "user_ip": ip_address,
            "continue_url": redirect_url,
            "target_project_id": project_id or self.project_id,
            "returnOobLink": not send_email,
        }
        response = self._execute(method=self.client.accounts().sendOobCode, body=data)

        if send_email:
            return {}

        url = response["oobLink"]
        query = parse_qs(urlparse(url).query)
        return {"url": url, "code": query["oobCode"][0]}

    def reset_password(self, email: str, new_password: str, old_password: str = None, oob_code: str = None):
        data = {
            "newPassword": new_password,
            "email": email,
        }

        if oob_code:
            data["oobCode"] = oob_code
        elif old_password:
            data["oldPassword"] = old_password
        else:
            raise exceptions.ValidationError("Either `old_password` or `oob_code` must be provided")

        response = self._execute(method=self.client.accounts().resetPassword, body=data)
        return response

    def delete_user(self, user_id: str):
        data = {
            "localId": user_id,
        }
        response = self._execute(method=self.client.accounts().delete, body=data)
        return response

    def disable_user(self, user_id: str, project_id: str = None):
        data = {
            "localId": user_id,
            "disableUser": True,
            "targetProjectId": project_id or self.project_id,
        }
        response = self._execute(method=self.client.accounts().update, body=data)
        return response

    def enable_user(self, user_id: str, project_id: str = None):
        data = {
            "localId": user_id,
            "disableUser": False,
            "targetProjectId": project_id or self.project_id,
        }
        response = self._execute(method=self.client.accounts().update, body=data)
        return response

    def sign_up(
        self,
        email: str,
        password: str,
        phone_number: str = None,
        name: str = None,
        photo_url: str = None,
        user_id: str = None,
        project_id: str = None,
    ):
        data = {
            "email": email,
            "password": password,
            "displayName": name,
            "photoUrl": photo_url,
            "disabled": False,
            "localId": user_id,
            "phoneNumber": phone_number,
            "targetProjectId": project_id or self.project_id,
        }
        response = self._execute(method=self.client.accounts().signUp, body=data)
        return User.create(data=response)

    def update(
        self,
        user_id: str,
        email: str = None,
        password: str = None,
        phone_number: str = None,
        name: str = None,
        photo_url: str = None,
        project_id: str = None,
        attributes: Dict[str, str] = None,
    ):
        data = {
            "localId": user_id,
            "targetProjectId": project_id or self.project_id,
        }
        if email:
            data["email"] = email
        if password:
            data["password"] = password
        if phone_number:
            data["phoneNumber"] = phone_number
        if name:
            data["displayName"] = name
        if photo_url:
            data["photoUrl"] = photo_url
        if attributes:
            data["customAttributes"] = json.dumps(attributes)

        response = self._execute(method=self.client.accounts().update, body=data)
        return response
